from colorama import Fore

from .cli import colourise

TAG_ORDER = [
    "workforce",
    "maintenance",
    "compliance",
    "customer",
    "internal",
    "sandbox",
    "production",
    "enterprise",
    "dedicated",
    "preview",
    "onboardng",
    "fast_release",
    "slow_release",
    "inactive",
]

TAG_COLOURS = {
    "workforce": Fore.LIGHTGREEN_EX,
    "maintenance": Fore.LIGHTMAGENTA_EX,
    "compliance": Fore.BLUE,
    "customer": Fore.LIGHTBLUE_EX,
    "internal": Fore.LIGHTCYAN_EX,
    "sandbox": Fore.LIGHTYELLOW_EX,
    "production": Fore.LIGHTRED_EX,
    "enterprise": Fore.WHITE,
    "dedicated": Fore.WHITE,
    "fast_release": Fore.YELLOW,
    "slow_release": Fore.MAGENTA,
    "inactive": Fore.RED,
}


def validate_tags(tags, other_valid_tags):
    unrecognised_tags = set(tags) - set(TAG_COLOURS.keys()) - set(other_valid_tags)
    if unrecognised_tags:
        raise Exception(f"Unrecognised tags: {', '.join(unrecognised_tags)}")


def colour_tag(tag):
    try:
        return colourise(tag, TAG_COLOURS[tag])
    except KeyError:
        return colourise(tag, Fore.LIGHTBLACK_EX)


def colour_tags(tags):
    return ", ".join(map(colour_tag, tags))


def sort_tags(tags):
    result = []
    for t in TAG_ORDER:
        if t in tags:
            result.append(t)
            tags.remove(t)
    return result + sorted(tags)



import os
import sys
from pathlib import Path

from . import monkeypatches  # NOQA
from .utils.cli import success, warning

__version__ = "0.9.16"


# Checking gitops version matches cluster repo version.
versions_path = (
    Path(os.environ.get("GITOPS_APPS_DIRECTORY", "apps")).parent.absolute() / "setup.cfg"
)
if versions_path.exists():
    import configparser

    from pkg_resources.extern.packaging.version import parse as parse_version

    config = configparser.RawConfigParser()
    config.read(versions_path)
    if "gitops.versions" in config:
        min_gitops_version = config.get("gitops.versions", "gitops")
        if parse_version(__version__) < parse_version(min_gitops_version):
            print(warning("Please upgrade Gitops."), file=sys.stderr)
            print(
                f"Your current version {success(__version__)} is less than the clusters minimum"
                f" requirement {success(min_gitops_version)}.",
                file=sys.stderr,
            )



import sys

from .cli import warning


class AppOperationAborted(Exception):
    pass


class AppDoesNotExist(Exception):
    def __init__(self, app_name=None):
        # Funky hack to stomp on traceback.
        if app_name:
            message = f"There's no app with the name '{app_name}', silly."
        else:
            message = "Could not find an 'apps' directory. Are you in a cluster repo?"
        self.args = (warning(message),)
        sys.exit(self)


class CommandError(Exception):
    pass



"""
REMOVE WHEN: https://github.com/pyinvoke/invoke/commit/8f6c0617c7dc59b105dd1b92fb417e75adc21bea is released.
"""
import inspect

if not hasattr(inspect, "getargspec"):
    inspect.getargspec = inspect.getfullargspec



import asyncio
import base64
import datetime
import os
import random
import time

import boto3
import dsnparse
from invoke import run, task

from .utils import kube
from .utils.apps import get_app_details
from .utils.cli import progress, success, warning

# TODO: Abstract this code out of gitops.


#####################
# DATABASE COMMANDS #
#####################
@task
def backup(ctx, app_name):
    """Backup production or staging database."""
    values = {"name": f"{app_name}-backup", "app": app_name}
    app = get_app_details(app_name, load_secrets=False)
    asyncio.run(
        kube._run_job(
            "jobs/backup-job.yml",
            values,
            context=app.cluster,
            namespace=app.namespace,
            sequential=True,
        )
    )


@task
def list_backups(ctx, app):
    kube.list_backups("workforce", app)


@task
def restore_backup(ctx, app_name, index, cleanup=True):
    """Restore backed up database."""
    kube.confirm_database(app_name)
    backups = kube.get_backups("workforce", app_name)
    backup = backups[int(index) - 1]
    values = {
        "name": f"{app_name}-restore",
        "timestamp": backup[0],
        "app": app_name,
    }
    app = get_app_details(app_name, load_secrets=False)
    asyncio.run(
        kube._run_job(
            "jobs/restore-job.yml",
            values,
            context=app.cluster,
            namespace=app.namespace,
            cleanup=cleanup,
        )
    )


@task
def copy_db(ctx, source, destination, skip_backup=False, cleanup=True):
    """Copy database between apps."""
    source_app = get_app_details(source, load_secrets=True)
    destination_app = get_app_details(destination, load_secrets=True)
    kube.confirm_database(destination)
    values = {
        "name": f"copy-db-{source}-{destination}",
        "source": source,
        "destination": destination,
        "SOURCE_DATABASE_URL_ENCODED": source_app.values["secrets"]["DATABASE_URL"],
        "DESTINATION_DATABASE_URL_ENCODED": destination_app.values["secrets"]["DATABASE_URL"],
        "skip_backup": "skip" if skip_backup else "",
    }
    asyncio.run(
        kube._run_job(
            "jobs/copy-db-job.yml",
            values,
            context=source_app.cluster,
            namespace=source_app.namespace,
            cleanup=cleanup,
        )
    )
    print(progress("You may want to clear the redis cache now!"))
    print(progress(f"\t- gitops mcommand {destination} clear_cache"))

    print(progress(":: Running migrations"))
    run(f"gitops migrate {destination}", echo=True)


@task
def download_backup(ctx, app, index=None, path=None, datestamp=False):
    """Download production or staging backup."""
    if not index:
        backups = kube.get_backups("workforce", app)
        if not backups:
            print(warning(f"No backups found for {app}"))
            return
        index = len(backups)
        print(progress("No index specified. Downloading latest backup: " + backups[-1][3]))

    kube.download_backup("workforce", app, index, path, datestamp=datestamp)


@task
def proxy(
    ctx,
    app_name,
    local_port=None,
    bastion_instance_id=None,
    aws_availability_zone=None,
    background=False,
):
    """Creates a proxy to RDS. Can supply either the app name or a DSN

    Usage: gitops db.proxy app_name
    or     gitops db.proxy postgres://...:...@5432/db
    """

    try:
        database_url = app_name
        database_dsn = dsnparse.parse(database_url)
        app_name = database_dsn.user
    except ValueError:
        # app_name was not a valid dsn
        app = get_app_details(app_name, load_secrets=True)
        database_url = base64.b64decode(
            app.values["secrets"]["DATABASE_URL"].encode("ascii")
        ).decode("ascii")
        database_dsn = dsnparse.parse(database_url)

    if not local_port:
        local_port = random.randint(1000, 9999)

    # Maybe we need to connect via RDS IAM instead!
    modified_dsn = dsnparse.parse(database_url)
    modified_dsn.hostname = "localhost"
    modified_dsn.port = local_port
    rds_iam = False

    if not database_dsn.password:
        result = run(
            f"aws rds generate-db-auth-token --hostname {database_dsn.host} --port"
            f" {database_dsn.port} --username {database_dsn.user}",
            hide=True,
        )
        modified_dsn.password = result.stdout.strip()
        rds_iam = True

    bastion_instance_id = bastion_instance_id or os.environ.get("GITOPS_BASTION_INSTANCE_ID")
    if not bastion_instance_id:
        raise Exception(
            "Please set GITOPS_BASTION_INSTANCE_ID environment variable for db proxy to work."
        )

    aws_availability_zone = aws_availability_zone or os.environ.get("GITOPS_AWS_AVAILABILITY_ZONE")
    if not aws_availability_zone:
        raise Exception(
            "Please set GITOPS_AWS_AVAILABILITY_ZONE environment variable for db proxy to work."
        )

    print(progress(f"Creating a proxy to the RDS instance of: {app_name} "))

    if rds_iam:
        proxy_dsn = (
            f"'user={modified_dsn.user} password={modified_dsn.password} host={modified_dsn.hostname} dbname={modified_dsn.dbname}"
            f" port={modified_dsn.port}'"
        )
    else:
        proxy_dsn = modified_dsn.geturl()
    print(
        progress(
            "Proxy established to :"
            f"postgres://{modified_dsn.user}@{modified_dsn.hostname}:{modified_dsn.port}/{modified_dsn.dbname}\n"
        )
    )
    print(progress(f"Connect to the db using: {proxy_dsn}\n"))
    cmd = f"""aws ssm start-session  \
        --target {bastion_instance_id}  \
        --document-name AWS-StartPortForwardingSessionToRemoteHost \
        --region {aws_availability_zone} \
        --parameters='{{"host": ["{database_dsn.host}"], "portNumber":["{database_dsn.port}"],"localPortNumber":["{local_port}"]}}'
    """
    return proxy_dsn, run(cmd, hide=True, asynchronous=background)


@task
def pgcli(
    ctx,
    app_name,
):
    """Opens pgcli to a remote DB"""
    print("making proxy")
    proxy_dsn, ctx = proxy(ctx, app_name, background=True)
    with ctx:
        print("Waiting for proxy to open")
        time.sleep(4)
        run(f"pgcli {proxy_dsn}", pty=True)


@task
def logs(ctx, app_name, last=24):
    """
    Fetches RDS logs since the last N hours.

    Usage:
        gitops db.logs APP --last 5
    """
    rds = boto3.client("rds")
    app = get_app_details(app_name, load_secrets=True)
    db_name = app.name.replace("_", "")
    logs = rds.describe_db_log_files(DBInstanceIdentifier=db_name)
    for log in logs["DescribeDBLogFiles"][-last:]:
        log_file = rds.download_db_log_file_portion(
            DBInstanceIdentifier=db_name,
            LogFileName=log["LogFileName"],
        )
        for line in log_file["LogFileData"].split("\n"):
            if line:
                try:
                    print(
                        str(
                            datetime.datetime.fromisoformat(line[:19])
                            .replace(tzinfo=datetime.timezone.utc)
                            .astimezone()
                        )
                        + line[23:]
                    )
                except Exception:
                    print(line)


@task
def wipe_db(ctx, destination, skip_backup=False, cleanup=True):
    """Wipes a customers database."""
    source_app = get_app_details(destination, load_secrets=True)
    kube.confirm_database(destination)
    values = {
        "name": f"wipe-db-{destination}",
        "destination": destination,
        "SOURCE_DATABASE_URL_ENCODED": source_app.values["secrets"]["DATABASE_URL"],
        "DESTINATION_DATABASE_URL_ENCODED": source_app.values["secrets"]["DATABASE_URL"],
        "skip_backup": "skip" if skip_backup else "",
    }
    print(progress(f":: Wiping database {destination}"))
    asyncio.run(
        kube._run_job(
            "jobs/wipe-db-job.yml",
            values,
            context=source_app.cluster,
            namespace=source_app.namespace,
            cleanup=cleanup,
        )
    )
    print(success("Deleted!"))
    print(progress(":: Running migrations"))
    run(f"gitops migrate {destination}", echo=True)

    print(progress(":: Clearing cache"))
    run(f"gitops mcommand {destination} clear_cache", echo=True)

    print(progress("Success!"))



import asyncio
import uuid

from colorama import Fore
from invoke import run, task

from .utils import config
from .utils.apps import APPS_PATH, get_apps, update_app
from .utils.async_runner import run_tasks_async_with_progress
from .utils.cli import colourise, progress, success, success_negative, warning
from .utils.exceptions import AppOperationAborted
from .utils.images import colour_image, get_image, get_latest_image
from .utils.kube import run_job
from .utils.tags import colour_tag, sort_tags


@task
def summary(ctx, filter="", exclude=""):
    """Produce a summary of apps, their tags, and their expected images & replicas.
    May not necessarily reflect actual app statuses if recent changes haven't yet been pushed to
    the remote, or the deployment has failed.
    """
    get_apps(
        filter=filter,
        exclude=exclude,
        mode="PREVIEW",
        autoexclude_inactive=False,
        load_secrets=False,
    )


@task
def bump(
    ctx,
    filter,
    exclude="",
    image_tag=None,
    prefix=None,
    autoexclude_inactive=True,
    interactive=True,
    push=False,
    redeploy=False,
    skip_migrations=False,
):
    """Bump image tag on selected app(s).
    Provide `image_tag` to set to a specific image tag, or provide `prefix` to use latest image
    with the given prefix.
    Otherwise, the latest tag with the same prefix as the app's current tag will be used.
    Provide `push` to automatically push the commit (and retry on conflict.)
    Provide `redeploy` to redeploy servers even if nothing has changed.
    Provide `skip_migrations` to disable running migrations via helm hooks.
    """
    prompt_message = "The following apps will have their image bumped"
    if image_tag:
        prompt_message += f" to use {colourise(image_tag, Fore.LIGHTYELLOW_EX)}"
    if prefix:
        prompt_message += f" to use prefix {colourise(prefix, Fore.LIGHTYELLOW_EX)}"
    try:
        apps = get_apps(
            filter=filter,
            exclude=exclude,
            autoexclude_inactive=autoexclude_inactive,
            message=f"{prompt_message}{colourise(':', Fore.LIGHTBLUE_EX)}",
            load_secrets=False,
            mode="PROMPT" if interactive else "SILENT",
        )
    except AppOperationAborted:
        print(success_negative("Aborted."))
        return

    if push:
        run(f"cd {APPS_PATH}; git pull")

    for app in apps:
        app_name = app.name
        prev_image_tag = app.image_tag
        if image_tag is None:
            # if we haven't specified a prefix
            if prefix is None:
                new_image_prefix = app.image_prefix
            else:
                new_image_prefix = prefix
            new_image_tag = get_latest_image(app.image_repository_name, new_image_prefix)
        else:
            new_image_tag = get_image(image_tag)
        if new_image_tag != prev_image_tag:
            print(
                f"Bumping {colourise(app_name, Fore.LIGHTGREEN_EX)}: {colour_image(prev_image_tag)}"
                f" -> {colour_image(new_image_tag)}"
            )
            update_app(app_name, **{"image-tag": new_image_tag})
        elif redeploy:
            print(f"Redeploying {colourise(app_name, Fore.LIGHTGREEN_EX)}")
            update_app(app_name, **{"bump": str(uuid.uuid4())})
        else:
            print(
                f"Skipping {colourise(app_name, Fore.LIGHTGREEN_EX)}: already on"
                f" {colour_image(new_image_tag)}"
            )
    if redeploy:
        commit_message = f"Redeploying {filter}"
    else:
        commit_message = f"Bump {filter}"

    if exclude:
        commit_message += f" (except {exclude})"
    if image_tag:
        commit_message += f" to use {image_tag}"
    if prefix:
        commit_message += f" to use prefix {prefix}"
    if skip_migrations:
        commit_message += " --skip-migrations"

    run(f'cd {APPS_PATH}; git commit -am "{commit_message}."')

    if push:
        git_push(APPS_PATH)

    print(success("Done!"))


@task
def command(
    ctx,
    filter,
    command,
    exclude="",
    cleanup=True,
    sequential=True,
    interactive=True,
    cpu=0,
    memory=0,
):
    """Run command on selected app(s).

    eg. inv command customer,sandbox -e aesg "python manage.py migrate"
    """
    try:
        apps = get_apps(
            filter=filter,
            exclude=exclude,
            message=(
                f"{colourise('The command', Fore.LIGHTBLUE_EX)}"
                f" {colourise(command, Fore.LIGHTYELLOW_EX)}"
                f" {colourise('will be run on the following apps:', Fore.LIGHTBLUE_EX)}"
            ),
            mode="PROMPT" if interactive else "SILENT",
        )
    except AppOperationAborted:
        print(success_negative("Aborted."))
        return

    # async output is by nature interactive
    if sequential or (not interactive) or len(apps) == 1:
        for app in apps:
            # For each app, just run the coroutine and print the output
            print(
                asyncio.run(
                    run_job(app, command, cleanup=cleanup, sequential=True, cpu=cpu, memory=memory)
                )
            )
    else:
        # Build list of coroutines, and execute them all at once
        jobs = [
            (
                run_job(
                    app, command, cleanup=cleanup, sequential=sequential, cpu=cpu, memory=memory
                ),
                app.name,
            )
            for app in apps
        ]
        asyncio.run(run_tasks_async_with_progress(jobs, max_concurrency=10))

    print(success("Done!"))


@task
def tag(ctx, filter, tag, exclude=""):
    """Set a tag on selected app(s)."""
    try:
        apps = get_apps(
            filter=filter,
            exclude=exclude,
            autoexclude_inactive=False,
            message=(
                f"{colourise('The tag', Fore.LIGHTBLUE_EX)} {colour_tag(tag)}"
                f" {colourise('will be added to the following apps:', Fore.LIGHTBLUE_EX)}"
            ),
            load_secrets=False,
        )
    except AppOperationAborted:
        print(success_negative("Aborted."))
        return
    for app in apps:
        update_app(app.name, tags=sort_tags(set(app.tags) | {tag}))
    commit_message = f"Add tag '{tag}' to {filter}"
    if exclude:
        commit_message += f" (except {exclude})"
    run(f'git commit -am "{commit_message}."')
    print(success("Done!"))


@task
def untag(ctx, filter, tag, exclude=""):
    """Unset a tag from selected app(s)."""
    try:
        apps = get_apps(
            filter=filter,
            exclude=exclude,
            autoexclude_inactive=False,
            message=(
                f"{colourise('The tag', Fore.LIGHTBLUE_EX)} {colour_tag(tag)}"
                f" {colourise('will be removed from the following apps:', Fore.LIGHTBLUE_EX)}"
            ),
            load_secrets=False,
        )
    except AppOperationAborted:
        print(success_negative("Aborted."))
        return
    for app in apps:
        update_app(app.name, tags=sort_tags(set(app.tags) - {tag}))
    commit_message = f"Remove tag '{tag}' from {filter}"
    if exclude:
        commit_message += f" (except {exclude})"
    run(f'git commit -am "{commit_message}."')
    print(success("Done!"))


@task  # TODO: want `keys` to be optional-positional: https://github.com/pyinvoke/invoke/issues/159
def getenv(ctx, filter, keys="", exclude=""):
    """Get one or more env vars on selected app(s)."""
    _getenv("environment", filter, exclude, keys)


@task  # TODO: want `keys` to be optional-positional: https://github.com/pyinvoke/invoke/issues/159
def getsecrets(ctx, filter, keys="", exclude=""):
    """Get one or more secrets on selected app(s)."""
    _getenv("secrets", filter, exclude, keys)


def _getenv(env_or_secrets, filter, exclude, filter_values):
    filter_values = filter_values.split(",") if filter_values else ""
    apps = get_apps(filter=filter, exclude=exclude, mode="SILENT")
    for app in apps:
        print("-" * 20, progress(app.name), sep="\n")
        values = app.values.get(env_or_secrets)
        if type(values) == dict:
            filtered_values = (
                {k: v for k, v in values.items() if k in filter_values} if filter_values else values
            )
            for k, v in filtered_values.items():
                print(f"{k}={v}")
        else:
            print(warning(f"No {env_or_secrets} set."))


def _sort_envs(envs):
    sorted_envs = {}
    for e in config.getlist("env_order", fallback=""):
        if e in envs:
            sorted_envs[e] = envs.pop(e)
    for e in sorted(envs):
        sorted_envs[e] = envs[e]
    return sorted_envs


@task
def setenv(ctx, filter, values, exclude=""):
    """Set one or more env vars on selected app(s).

    eg. inv setenv customer,sandbox BG_RUNNER=DRAMATIQ,BUMP=2

    NOTE: More broad-reaching environment changes should be made at the chart level.
    """
    splitenvs = values.split(",")  # pardon the pun.
    formatted_splitenvs = "\n".join(splitenvs)
    try:
        apps = get_apps(
            filter=filter,
            exclude=exclude,
            message=(
                f"{colourise('The env var(s)', Fore.LIGHTBLUE_EX)}\n{colourise(formatted_splitenvs, Fore.LIGHTYELLOW_EX)}\n{colourise('will be added to the following apps:', Fore.LIGHTBLUE_EX)}"
            ),
        )
    except AppOperationAborted:
        print(success_negative("Aborted."))
        return
    for app in apps:
        update_app(
            app.name,
            environment=_sort_envs(
                {
                    **dict(tuple(e.split("=")) for e in splitenvs),
                    **app.values.get("environment", {}),
                }
            ),
        )
    commit_message = f"Set env var(s) '{values}' on {filter}"
    if exclude:
        commit_message += f" (except {exclude})"
    run(f'git commit -am "{commit_message}."')
    print(success("Done!"))


@task
def unsetenv(ctx, filter, values, exclude=""):
    """Unset one or more env vars on selected app(s).

    eg. inv unsetenv customer,sandbox BG_RUNNER,BUMP

    NOTE: More broad-reaching environment changes should be made at the chart level.
    """
    splitenvs = values.split(",")  # pardon the pun.
    formatted_splitenvs = "\n".join(splitenvs)
    try:
        apps = get_apps(
            filter=filter,
            exclude=exclude,
            message=(
                f"{colourise('The env var(s)', Fore.LIGHTBLUE_EX)}\n{colourise(formatted_splitenvs, Fore.LIGHTYELLOW_EX)}\n{colourise('will be removed from the following apps:', Fore.LIGHTBLUE_EX)}"
            ),
        )
    except AppOperationAborted:
        print(success_negative("Aborted."))
        return
    for app in apps:
        environment = app.values.get("environment", {})
        for e in splitenvs:
            if e in environment:
                del environment[e]
        update_app(app.name, environment=_sort_envs(environment))
    commit_message = f"Unset env var(s) '{values}' on {filter}"
    if exclude:
        commit_message += f" (except {exclude})"
    run(f'git commit -am "{commit_message}."')
    print(success("Done!"))


@task
def setcluster(ctx, filter, cluster, exclude=""):
    """Move selected app(s) to given cluster.

    eg. inv setcluster customer,sandbox eks-prod
    """
    try:
        apps = get_apps(
            filter=filter,
            exclude=exclude,
            message=(
                f"{colourise('The following apps will be moved to the', Fore.LIGHTBLUE_EX)}"
                f" {colourise(cluster, Fore.LIGHTYELLOW_EX)}"
                f" {colourise('cluster:', Fore.LIGHTBLUE_EX)}"
            ),
        )
    except AppOperationAborted:
        print(success_negative("Aborted."))
        return
    for app in apps:
        update_app(app.name, cluster=cluster)
    commit_message = f"Move {filter} to cluster '{cluster}'"
    if exclude:
        commit_message += f" (except {exclude})"
    run(f'git commit -am "{commit_message}."')
    print(success("Done!"))


def git_push(cluster_path: str, retry: int = 3):
    """Git pushes in a directory and retries if commits already exist"""
    print(progress(f"Pushing changes to {cluster_path}"))
    attempts = 0
    while attempts <= retry:
        result = run(f"cd {cluster_path}; git push", warn=True)
        if result.exited == 1 and "remote contains work" in result.stderr:
            attempts += 1
            run(f"cd {cluster_path}; git pull --rebase=true")
        else:
            break



import json
import os
from base64 import b64encode
from typing import Dict, List, Optional, Union

from .utils import load_yaml

DEPLOYMENT_ATTRIBUTES = [
    "tags",
    "image-tag",
    "containers",
    "environment",
    "cluster",
    "bump",
    "namespace",
]


class App:
    def __init__(
        self,
        name: str,
        path: Optional[str] = None,
        deployments: Optional[Dict] = None,
        secrets: Optional[Dict] = None,
        load_secrets: bool = True,
        account_id: str = "",
    ):
        self.name = name
        self.path = path
        self.account_id = account_id
        if path:
            deployments = load_yaml(os.path.join(path, "deployment.yml"))
            if load_secrets:
                secrets = load_yaml(os.path.join(path, "secrets.yml")).get("secrets", {})
        deployments = deployments or {}
        secrets = secrets or {}
        self.values = self._make_values(deployments, secrets)
        self.namespace: str = self.values["namespace"]
        self.chart = Chart(self.values["chart"])

    def __eq__(self, other):
        return (
            type(self) == type(other)
            and self.name == other.name
            and json.dumps(self.values, sort_keys=True) == json.dumps(other.values, sort_keys=True)
        )

    def is_inactive(self):
        return "inactive" in self.values.get("tags", [])

    def set_value(self, path: str, value: any):
        """Sets the value on the path of self.values
        Usage:
            app.set_value("deployments.label.GITHUB_DEPLOYMENT_KEY", "1")
        """
        keys = path.split(".")
        current_dict = self.values
        for key in keys[:-1]:
            current_dict = current_dict.setdefault(key, {})
        current_dict[keys[-1]] = value

    def _make_values(self, deployments: Dict, secrets: Dict) -> Dict:
        values = {
            **deployments,
            "secrets": {**{k: b64encode(v.encode()).decode() for k, v in secrets.items()}},
        }

        image = self._make_image(deployments)
        if image:
            values["image"] = image

        # Don't include the `images` key. It will only cause everything to be
        # redeployed when any group changes.
        values.pop("images", None)
        return values

    def _make_image(self, deployment_config: Dict):
        if "image-tag" in deployment_config:
            return deployment_config["images"]["template"].format(
                account_id=self.account_id,
                tag=deployment_config["image-tag"],
            )
        else:
            return deployment_config.get("image", "")

    @property
    def image(self) -> str:
        """[305686791668.dkr.ecr.ap-southeast-2.amazonaws.com/uptick:yoink-9f03ac80f3]"""
        image = self.values.get("image", "")
        if isinstance(image, dict):
            return f"{image['repository']}:{image.get('tag','latest')}"
        else:
            return image

    @property
    def image_repository_name(self) -> str:
        """305686791668.dkr.ecr.ap-southeast-2.amazonaws.com/[uptick]:yoink-9f03ac80f3"""
        return self.image.split(":")[0].split("/")[-1]

    @property
    def image_tag(self) -> str:
        """305686791668.dkr.ecr.ap-southeast-2.amazonaws.com/uptick:[yoink-9f03ac80f3]"""
        return self.image.split(":")[-1]

    @property
    def image_prefix(self) -> str:
        """Gets the image prefix portion of {prefix}-{git hash}
        305686791668.dkr.ecr.ap-southeast-2.amazonaws.com/uptick:[yoink]-9f03ac80f3

        eg: qa-server-12345 -> qa-server
        """
        if "-" not in self.image_tag:
            return ""
        return self.image_tag.rsplit("-", 1)[0]

    @property
    def cluster(self) -> str:
        return self.values.get("cluster", "")

    @property
    def tags(self) -> List[str]:
        return self.values.get("tags", [])

    @property
    def service_account_name(self) -> str:
        return (
            self.values.get("serviceAccount", {}).get("name")
            or self.values.get("serviceAccountName")
            or "default"
        )


class Chart:
    """Represents a Helm chart

    Can be stored in a git repo, helm repo or local path

    Example definition in `deployments.yml`:
    chart:
      type: git
      git_sha: develop
      git_repo_url: https://github.com/uptick/workforce

    of
    chart:
      type: helm
      helm_repo: brigade
      helm_repo_url: https://brigadecore.github.io/charts
      helm_chart: brigade/brigade

    If a dictionary is not passed, it is assumed to be a git repo. eg:
      chart: https://github.com/uptick/workforce
    """

    def __init__(self, definition: Union[Dict, str]):
        if isinstance(definition, str):
            # for backwards compat, any chart definition which is a string, is a git repo
            self.type = "git"
            self.git_sha = None
            self.git_repo_url = definition
        elif isinstance(definition, dict):
            self.type = definition["type"]
            self.git_sha = definition.get("git_sha")
            self.git_repo_url = definition.get("git_repo_url")
            self.helm_repo = definition.get("helm_repo")
            self.helm_repo_url = definition.get("helm_repo_url")
            self.helm_chart = definition.get("helm_chart")
            self.version = definition.get("version")
            self.path = definition.get("path")
        else:
            raise Exception(
                "Chart definition must be either a dict or string. Instead it is: {definition}"
            )

        if self.git_repo_url and "@" in self.git_repo_url:
            self.git_repo_url, self.git_sha = self.git_repo_url.split("@")



from functools import lru_cache
from hashlib import md5

import boto3
from colorama import Fore

from .cli import colourise

BATCH_SIZE = 100


def get_image(tag):
    """Finds a specific image in ECR."""
    # TODO
    raise NotImplementedError


@lru_cache
def get_latest_image(repository_name: str, prefix: str) -> str:
    """Finds latest image in ECR with the given prefix and returns the image tag"""
    ecr_client = boto3.client("ecr")
    client_paginator = ecr_client.get_paginator("describe_images")

    results = []
    for ecr_response in client_paginator.paginate(
        repositoryName=repository_name,
        filter={"tagStatus": "TAGGED"},
        maxResults=BATCH_SIZE,
    ):
        for image in ecr_response["imageDetails"]:
            if prefix != "":
                if prefix_tags := [
                    tag for tag in image["imageTags"] if tag.startswith(prefix + "-")
                ]:
                    results.append((prefix_tags[0], image["imagePushedAt"]))
            else:
                if prefix_tags := [tag for tag in image["imageTags"] if "-" not in tag]:
                    results.append((prefix_tags[0], image["imagePushedAt"]))

    if not results:
        if prefix:
            print(f'No images found in repository: {repository_name} with tag "{prefix}-*".')
        else:
            print(f"No images found in repository: {repository_name}")
        return None

    latest_image_tag = sorted(results, key=lambda image: image[1], reverse=True)[0][0]
    return latest_image_tag


def colour_image(image_tag: str) -> str:
    if not image_tag:
        return image_tag

    bits = image_tag.split("-")
    if len(bits) > 1:
        bits[0] = colourise(bits[0], color_hash(bits[1]))
        return "-".join(bits)
    else:
        return colourise(bits[0], color_hash(bits[0]))


def color_hash(bit):
    return [
        Fore.RED,
        Fore.GREEN,
        Fore.YELLOW,
        Fore.BLUE,
        Fore.MAGENTA,
        Fore.CYAN,
        Fore.WHITE,
    ][int.from_bytes(md5(bit.encode()).digest(), "big") % 7]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gitops', 'gitops.common', 'gitops.utils']

package_data = \
{'': ['*']}

install_requires = \
['boto3',
 'colorama>=0.4.4,<0.5.0',
 'dsnparse',
 'humanize>=3.5.0,<4.0.0',
 'invoke',
 'pyyaml>=5.1.2',
 'tabulate>=0.8.9,<0.9.0']

extras_require = \
{'server': ['fastapi',
            'httpx>=0.18.1,<0.24.0',
            'uvicorn>=0.17.0,<0.18.0',
            'uvicorn>=0.17.0,<0.18.0',
            'sentry-sdk>=1.3.0,<2.0.0',
            'kubernetes_asyncio>=12.1.2,<13.0.0']}

entry_points = \
{'console_scripts': ['gitops = gitops.main:program.run']}

setup_kwargs = {
    'name': 'gitops',
    'version': '0.9.16',
    'description': 'Manage multiple apps across one or more k8s clusters.',
    'long_description': 'None',
    'author': 'Jarek Głowacki',
    'author_email': 'jarekwg@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
